%% Code for Day 5

% Add path to EEGLab and launch EEGLab before we begin (or use your
% shortcut on the shortcuts bar)

% Make sure you can access your sample data (sampleEEGdata.mat)


%% -------------------------------------------
% REVIEW and GETTING TO KNOW YOUR SAMPLE DATA
% --------------------------------------------

%% ASSIGNMENT: Time Domain Averaging and Plotting --------------

% load your sample data

% Making ERPs and plotting
% * Compute the ERP at each electrode. 
ERP_data = @@ % You can do this!

% * Create butterfly plot of all channels, with time as X axis in msec

% * In a new figure, plot only channel Pz

% * Select five time points at which to show topographical plots (e.g., 0 to 400 ms in 100-ms steps). 
%     See help topoplot, sample below
time2plot_idx = @@   % note this is the index
topoplot(ERP_data(:,time2plot_idx),EEG.chanlocs);

% * In one figure, make a series of topographical plots at these time points. 
%   ** Indicate the time point in a title on each subplot. 


%% ASSIGNMENT: Wavelet convolution and plotting %% ------------

% Create a family of Morlet wavelets ranging in frequency from 2 Hz to 30 Hz
%   in five steps. 

% Select one electrode from the scalp EEG dataset and convolve each wavelet 
%   with EEG data from all trials from that electrode. 

% Apply the Matlab function real to the convolution result, 
%   as in convol_result=real(convol_result). This will return the 
%   EEG data bandpass filtered at the peak frequency of the wavelet. 

% Average the result of convolution over all trials and plot an ERP
%   corresponding to each wavelet frequency.  Each frequency 
%   should be in its own subplot, and leave room for a 6th subplot (for below). 

% Plot the broadband ERP (without any convolution) in the 6th subplot. 

% How do the wavelet-convolved ERPs compare with the broadband ERP? 
%   Are there dynamics revealed by the wavelet-convolved ERPs that are 
%   not apparent in the broadband ERP, and are there dynamics in the 
%   broadband ERP that are not apparent in the waveletconvolved ERPs? 
%   Base your answer on qualitative visual inspection of the results; 
%   statistics or other quantitative comparisons are not necessary.

% Select the same five time points you selected above for the time-domain ERP plotting
%   Plot the same five points for wavelet-convolved data (once for each frequency)
%   Make a plot with subplots, with rows corresponding to the subplots above, 
%   and columns to the timepoints, with a topoplot in each

%% -------------------------------------------------------------------
% More fun with Wavelets: Real Signal, Magnitude (Power), Phase Angle
% -------------------------------------------------------------------_
% Clear workspace and reload sample data
clear all
load('..\sampleEEGdata.mat'); % change directory as needed

%% create a Morlet wavelet
srate = EEG.srate; % in hz
time  = -2:1/srate:2; % best practice is to have time=0 at the center of the wavelet
frex  = 6.5; % frequency of wavelet, in Hz

% create complex sine wave
sine_wave = exp( 1i*2*pi*frex.*time );

% create Gaussian window
s = 7 / (2*pi*frex); % this is the standard deviation of the gaussian
gaus_win  = exp( (-time.^2) ./ (2*s^2) );

% now create Morlet wavelet
cmw  = sine_wave .* gaus_win;

% Plot real and imaginary part of wavelet
plot(time,real(cmw),'b')
hold on
plot(time,imag(cmw),'b:')
legend({'real part';'imaginary part'})

%% convolve this complex wavelet with the data

% Get one channel, one trial
chan2use  = 'fz';
trial2use = 34;
% extract this bit of data for convenience 
data = squeeze(EEG.data( strcmpi(chan2use,{EEG.chanlocs.labels}), :,trial2use));

% do the convolution using Mike's code that I made into a function
as = cmw_conv(data,cmw);

%% visualize the outputs of convolution

figure(1), clf
% plot the filtered signal (projection onto real axis)
s1 = subplot(311)
plot(EEG.times,real(as))
xlabel('Time (ms)'), ylabel('Amplitude (\muV)')
set(gca,'xlim',[-200 1300])


% plot power (squared magnitude from origin to dot-product location in
% complex space)
s2 = subplot(312)
plot(EEG.times,abs(as).^2)
xlabel('Time (ms)'), ylabel('Power \muV^2')
set(gca,'xlim',[-200 1300])


% plot phase (angle of vector to dot-product, relative to positive real
% axis)
s3 = subplot(313)
plot(EEG.times,angle(as))
xlabel('Time (ms)'), ylabel('Phase (rad.)')
set(gca,'xlim',[-200 1300])

% link axes
linkaxes([s1 s2 s3],'x');

%% -------------------------------------------
% More fun with Wavelets: That "s" parameter!
% --------------------------------------------
% Clear workspace and reload sample data
clear all
load('..\sampleEEGdata.mat'); % change directory as needed

% set a few different wavelet widths ("number of cycles" parameter)
num_cycles = [ 2 4 6 10 ];
frex = 6.4;
time = -2:1/EEG.srate:2;

figure(1), clf
%% Plot Gaussian vs time to show temporal resolution
for i=1:4
    subplot(4,2,2*i-1)
    s = num_cycles(i) / (2*pi*frex);
    plot(time, exp( (-time.^2) ./ (2*s^2) ),'k','linew',3)
    title([ 'Gaussian with ' num2str(num_cycles(i)) ' cycles' ])
end
xlabel('Time (s)'), ylabel('Amplitude')

%% show frequency spectrum of wavelets with different number of cycles
% plot freq response in second column (gain vs freq)
for i=1:4
    subplot(4,2,2*i)
    s = num_cycles(i) / (2*pi*frex);
    cmw = exp(1i*2*pi*frex.*time) .* exp( (-time.^2) ./ (2*s^2) );
    
    % take its FFT
    cmwX = fft(cmw);
    cmwX = cmwX./max(cmwX);
    hz = linspace(0,EEG.srate/2,floor(length(time)/2)+1);
    
    % plot it
    plot(hz,abs(cmwX(1:length(hz))),'k','linew',3)
    set(gca,'xlim',[0 20])
    title([ 'Power of wavelet with ' num2str(num_cycles(i)) ' cycles' ])
end
xlabel('Frequency (Hz)'), ylabel('Amplitude')

%% ASSIGNMENT: What happened to i?
% is i the imaginary number now?
% evaluate i
% how would you make i the imaginary number again?

%% ASSIGNMENT: Explore time-freq trade-off %% ------------
% Change the frequency for the code above and examine how the number of
% cycles affects frequencies in delta, theta, alpha, beta, and gamma
% Just change the frequency of the wavelet and rerun the code

%% ----------------------------------------------------------------------
% Time-frequency plotting and visualizing the impact of number of cycles 
% -----------------------------------------------------------------------
% clear the workspace and reload that data!
clear all
load('..\sampleEEGdata.mat'); % change directory as needed

% compare a fixed number of wavelet cycles (not variable -- see below)
num_cycles = [ 2 4 6 10 ];

% wavelet parameters for a wavelet "family"
num_frex = 40;
min_freq =  2;
max_freq = 30;

% let's use only one channel, OK?
channel2use = 'o1'; % choose your favorite channel


% other wavelet parameters
frex = linspace(min_freq,max_freq,num_frex);
time = -2:1/EEG.srate:2;
half_wave = (length(time)-1)/2;

% initialize output time-frequency data
tf = zeros(length(num_cycles),length(frex),EEG.pnts);

% Data reshaped across trials (doesn't change on frequency iteration)
data = reshape(EEG.data(strcmpi(channel2use,{EEG.chanlocs.labels}),:,:),1,[]);

% loop over cycles
for cyclei=1:length(num_cycles)
    
    for fi=1:length(frex)
        
        % create wavelet 
        disp(['Making your peronalized ' num2str(frex(fi)) ' hz wavelet with ' num2str(num_cycles(cyclei)) ' cycles']);
        s = num_cycles(cyclei)/(2*pi*frex(fi));
        cmw  = exp(2*1i*pi*frex(fi).*time) .* exp(-time.^2./(2*s^2));
        
        % get analytical signal
        as = cmw_conv(data,cmw);
        
        % and reshape back to time X trials
        as = reshape( as, EEG.pnts, EEG.trials );
                
        % put power data into big matrix
        tf(cyclei,fi,:) = mean(abs(as).^2,2);
    end

    % baseline for plotting purposes (to do db correction)
    baseline_window = [ -500 -200 ];
    % convert baseline time into indices
    [~,baseidx(1)] = min(abs(EEG.times-baseline_window(1)));
    [~,baseidx(2)] = min(abs(EEG.times-baseline_window(2)));
    % db conversion
    tf(cyclei,:,:) = 10*log10( bsxfun(@rdivide, squeeze(tf(cyclei,:,:)), mean(tf(cyclei,:,baseidx(1):baseidx(2)),3)' ) );
    
end

%% plot results
n_contour_levels=40;

figure(3), clf
for cyclei=1:length(num_cycles)
    subplot(3,2,cyclei)
    contourf(EEG.times,frex,squeeze(tf(cyclei,:,:)),n_contour_levels,'linecolor','none')
    set(gca,'clim',[-3 3],'ydir','normal','xlim',[-300 1000])
    title([ 'Wavelet with ' num2str(num_cycles(cyclei)) ' cycles' ])
    xlabel('Time (ms)'), ylabel('Frequency (Hz)')
end

%% ASSIGNMENT: Explore time-freq trade-off and contourf %% ------------
% What do these contour plots tell you about the t-f tradeoff and number of cycles?
% What number of cycles is optimal
% How do you use contourf?  What is the impact of changing the number of contour levels?

% Before continuing, change number of contour levels to 40 and rerun the plot above

%% Variable number of cycles that are frequency-dependent

% set a few different wavelet widths (number of wavelet cycles)
range_cycles = [ 3 10 ]; % You might choose others if you wish

% other wavelet parameters
nCycles = logspace(log10(range_cycles(1)),log10(range_cycles(end)),num_frex);

% initialize output time-frequency data
tf = zeros(length(frex),EEG.pnts);

for fi=1:length(frex)

    % create wavelet 
    disp(['Making your peronalized ' num2str(frex(fi)) ' hz wavelet with ' num2str(nCycles(fi)) ' cycles']);
    s = nCycles(fi)/(2*pi*frex(fi));
    cmw  = exp(2*1i*pi*frex(fi).*time) .* exp(-time.^2./(2*s^2));

    % get analytical signal
    as = cmw_conv(data,cmw);

    % and reshape back to time X trials
    as = reshape( as, EEG.pnts, EEG.trials );

    % put power data into big matrix
    tf(fi,:) = mean(abs(as).^2,2);
end


%% Now plot the variable cycle results
% baseline for plotting purposes (to do db correction)
baseline_window = [ -500 -200 ];
% convert baseline time into indices
[~,baseidx(1)] = min(abs(EEG.times-baseline_window(1)));
[~,baseidx(2)] = min(abs(EEG.times-baseline_window(2)));
% db conversion
tf = 10*log10( bsxfun(@rdivide, tf, mean(tf(:,baseidx(1):baseidx(2)),2)) );
% plot results
subplot(3,2,5)
contourf(EEG.times,frex,tf,n_contour_levels,'linecolor','none')
set(gca,'clim',[-3 3],'ydir','normal','xlim',[-300 1000])
title('Convolution with a range of cycles')
xlabel('Time (ms)'), ylabel('Frequency (Hz)')

%% --------------------------------------
% Filtering -- methods for design
% ---------------------------------------
% clear the workspace and reload that data!
clear all
load('..\sampleEEGdata.mat'); % change directory as needed

% specify Nyquist freuqency
nyquist = EEG.srate/2;

%% firls
% some parameters to define filter
% filter frequency band
filtbound = [4 10]; % Hz
% transition width
trans_width = 0.2; % fraction of 1, thus 20%
% filter order
filt_order = round(3*(EEG.srate/filtbound(1)));
% frequency vector (as fraction of Nyquist
ffrequencies  = [ 0 (1-trans_width)*filtbound(1) filtbound (1+trans_width)*filtbound(2) nyquist ]/nyquist;
% shape of filter (must be the same number of elements as frequency vector
idealresponse = [ 0 0 1 1 0 0 ];

%% Plot the ideal filter response
figure; plot(ffrequencies,idealresponse);

%% get filter weights using firls
filterweights_firls = firls(filt_order,ffrequencies,idealresponse);

%% use freqz to obtain gain function and plot it
[h,w] = freqz(filterweights_firls);
gain = abs(h);
f_axis = w/pi*nyquist;
figure('numbertitle','off','Name','Transfer function of firls versus ideal');
plot(ffrequencies*nyquist,idealresponse)
hold on;
plot(w/pi*nyquist,abs(h),'color',[0.5 0 1]);
xlabel 'Hz';
ylabel 'Gain';
ylim([0 1.2]);


%% fir1
normalized_filterbound = filtbound/nyquist;
filterweights_fir1 = fir1(filt_order,normalized_filterbound); 
xlabel 'Hz';
ylabel 'Gain';
ylim([0 1.2]);

%% use freqz to obtain gain function and plot it
[h,w] = freqz(filterweights_fir1);
gain = abs(h);
f_axis = w/pi*nyquist;
figure('numbertitle','off','Name','Transfer function of fir1 versus ideal');
plot(ffrequencies*nyquist,idealresponse)
hold on;
plot(w/pi*nyquist,abs(h),'color',[0.5 0 1]);
xlabel 'Hz';
ylabel 'Gain';
ylim([0 1.2]);

%% fir2 -- have more control over filter response 

% define frequency vector from 0 to 1, where 1 is nyquist
T = filt_order/EEG.srate;  % length of epoch in time
fs = 0:1/T:EEG.srate/2;     % frequency vector from 0 to Nyquist
f = fs./(EEG.srate/2); % normalized frequency vector from 0 to Nyquist 
% Define ideal frequency function
m = fs >= filtbound(1) & fs <= filtbound(2);   % where the frequency is between cutpoints
filterweights_fir2 = fir2(filt_order,f,m);


[h,w] = freqz(filterweights_fir2);
gain = abs(h);
f_axis = w/pi*nyquist;

figure('numbertitle','off','Name','Transfer function of FIR2 versus ideal');
plot(fs,m,'-k');
hold on;
plot(w/pi*nyquist,abs(h),'color',[0.5 0 1]);
xlabel 'Hz';
ylabel 'Gain';
ylim([0 1.2]);

%% Compare all three
figure('numbertitle','off','Name','Transfer function of FIRLS, FIR1, FIR2 versus ideal');

% ideal
plot(fs,m,'-k');
hold on;

% firls
[h,w] = freqz(filterweights_firls);
gain = abs(h);
f_axis = w/pi*nyquist;
plot(w/pi*nyquist,abs(h),'-b');

% fir1
[h,w] = freqz(filterweights_fir1);
gain = abs(h);
f_axis = w/pi*nyquist;
plot(w/pi*nyquist,abs(h),'-r');

% fir2
[h,w] = freqz(filterweights_fir2);
gain = abs(h);
f_axis = w/pi*nyquist;
plot(w/pi*nyquist,abs(h),'color',[0.5 0 1]);

legend({'ideal', 'firls','FIR1','FIR2'});
xlabel 'Hz';
ylabel 'Gain';
ylim([0 1.2]);

%% ASSIGNMENT: Wavelet frequency responses
% Make a family of wavelets and use the freqz function to 
% loop and plot the transfer functions on a single plot as we just did for
% FIR filters

%% ASSIGNMENT we'll likely not get to today
% Reshape the single channel of data to a vector
% Filter the data with a filter of your choosing
% Obtain the analytic signal using the hilbert function
% Reshape the analytic signal back to times by trials
% Take the mean power across epochs to have a vector of mean power by time
% In two subplots, plot the ERP and the bandpass-filtered hilbert power signal

